// ExeFunction.cpp
// Project template VS2008 2D images FDA format
// MM IFPEN 2018

#include "stdafx.h"

using namespace System::Drawing;
using namespace System::IO;
using namespace System::Collections::Generic;
using namespace System::Windows::Media::Imaging;
using namespace System;
using namespace System::Text::RegularExpressions;
using namespace System::Xml;

#define CODEREADSAVEFUC 100
#define CODEREADSAVEFI 101
#define CODEREADSAVEFF 102
#define CODEREADSAVEFD 103
#define CODEREADSAVEFCX 104
#define CODEREADSAVEF2UC 200
#define CODEREADSAVEF2I 201
#define CODEREADSAVEF2F 202
#define CODEREADSAVEF2D 203
#define CODEREADSAVEF2CX 204
#define CODEREADSAVEF2UCRGB 210
#define CODEREADSAVEF3UC 300
#define CODEREADSAVEF3I 301
#define CODEREADSAVEF3F 302
#define CODEREADSAVEF3D 303
#define CODEREADSAVEF3CX 304
#define CODEREADSAVEF3UCRGB 310

template <class CDaTa> class F2DTModel
{
  public:
  CDaTa *DT;
  int W;
  int H;

  private :
  int Wstride;

  public:
  F2DTModel<CDaTa>(): W(1),H(1),Wstride(1*sizeof(CDaTa)) {DT=new CDaTa[1]; DT[0]=0;}; 
  F2DTModel<CDaTa>(const F2DTModel<CDaTa> &A): W(A.W),H(A.H), Wstride(A.Wstride) {int T=W*H; DT=new CDaTa[T]; for (int u=0;u<T;u++) DT[u]=A.DT[u]; }
  F2DTModel<CDaTa>(int _W,int _H) { if (_W<=0) _W=1; if (_H<=0) _H=1; W=_W; H=_H; Wstride=W*sizeof(CDaTa); DT=new CDaTa[W*H];}
  ~F2DTModel<CDaTa>() {if (DT) delete[] DT;}
  CDaTa operator[] (int index) const {return DT[index];}
  CDaTa * operator[] (int index) {return DT+(Wstride*index);}
  CDaTa operator() (int y, int x) const {return DT[y*W+x];}
  CDaTa & operator() (int y,int x) {return DT[y*W+x];}
  CDaTa operator() (int index) const {return DT[index];}
  CDaTa & operator() (int index) {return DT[index];}
	F2DTModel<CDaTa>& operator=(const F2DTModel<CDaTa>& A) { if (this != &A) { delete[] DT; W = A.W; H = A.H; Wstride = A.Wstride; int T = W*H; DT = new CDaTa[T]; /*memcpy((void*)DT, (void*)A.DT, sizeof(DT));*/ for (int u=0;u<T;u++) DT[u]=A.DT[u]; } return (*this); }
  int GetW() {return W;}
  int GetH() {return H;}
  int GetWstride() {return Wstride;}
  void SetWstride(int W) {Wstride=W*sizeof(CDaTa);}
};

template <class CDaTa> int GetW(F2DTModel<CDaTa> &I) {return I.W;}
template <class CDaTa> int GetH(F2DTModel<CDaTa> &I) {return I.H;}

typedef  F2DTModel<unsigned char> F2UC;
typedef  F2DTModel<int> F2I;
typedef  F2DTModel<float> F2F;
typedef  F2DTModel<double> F2D;

void Resize(F2UC &T,int W,int H) {T.W=W; T.H=H; T.SetWstride(W); delete[] T.DT; T.DT=new unsigned char[W*H];}
void Resize(F2I &T,int W,int H) {T.W=W; T.H=H; T.SetWstride(W); delete[] T.DT; T.DT=new int[W*H];}
void Resize(F2F &T,int W,int H) {T.W=W; T.H=H; T.SetWstride(W); delete[] T.DT; T.DT=new float[W*H];}
void Resize(F2D &T,int W,int H) {T.W=W; T.H=H; T.SetWstride(W); delete[] T.DT; T.DT=new double[W*H];}

template <class CDaTa> void Create(F2DTModel<CDaTa> &T,int W,int H) {Resize(T,W,H);}

int GetFormat(F2UC &A) { return CODEREADSAVEF2UC; }
int GetFormat(F2I &A) { return CODEREADSAVEF2I; }
int GetFormat(F2F &A) { return CODEREADSAVEF2F; }
int GetFormat(F2D &A) { return CODEREADSAVEF2D; }

bool LoadF2UC(String ^fn, F2UC &R)
{
	FileStream^ fs;
	try
	{
		if (!File::Exists(fn))
			return false;

		int Header[3];
		fs = gcnew FileStream(fn, FileMode::Open);
		BinaryReader^ wb = gcnew BinaryReader(fs);

		for (int i = 0; i<3; i++)
		{
			Header[i] = wb->ReadInt32();
		}

		if (Header[0] != CODEREADSAVEF2UC)
			return false;

		int WDaTa = Header[1];
		int HDaTa = Header[2];
		int TDaTa = WDaTa*HDaTa;
		Resize(R, WDaTa, HDaTa);

		for (int i = 0; i<TDaTa; i++)
			R(i) = wb->ReadByte();

		fs->Close();
		return true;
	}
	catch (System::Exception ^ex) 
	{ if (fs) fs->Close(); return false; }
	return false;
}


bool LoadRGBF2UC(String ^fn, F2UC &IR, F2UC &IG, F2UC &IB, F2UC &I)
{
	FileStream^ fs;
	try
	{
		if (!File::Exists(fn))
			return false;
		int Header[3];
		fs = gcnew FileStream(fn, FileMode::Open);
		BinaryReader^ wb = gcnew BinaryReader(fs);
		for (int i = 0; i<3; i++)
  		  Header[i] = wb->ReadInt32();
		if (Header[0] != CODEREADSAVEF2UCRGB)
			return false;
		int WDaTa = Header[1];
		int HDaTa = Header[2];
		int TDaTa = WDaTa*HDaTa;
		Create(IR, WDaTa, HDaTa);
		Create(IG, WDaTa, HDaTa);
		Create(IB, WDaTa, HDaTa);
		Create(I, WDaTa, HDaTa);
		for (int i = 0; i<TDaTa; i++)
			IR(i) = wb->ReadByte();
		for (int i = 0; i<TDaTa; i++)
			IG(i) = wb->ReadByte();
		for (int i = 0; i<TDaTa; i++)
			IB(i) = wb->ReadByte();
		int val;
		for (int u = 0; u<TDaTa; u++)
		{
			val = IR(u) + IG(u) + IB(u);
			I(u) = (val / 3);
		}
		fs->Close();
		return true;
	}
	catch (System::Exception ^ex) { if (fs) fs->Close(); return false; }
	return false;
}


bool SaveF2UC(F2UC &I, String ^fn)
{
	FileStream^ fs;
	try
	{
		int code = CODEREADSAVEF2UC;
		int Header[3] = { code, I.W, I.H };
		int T = I.W*I.H;

		fs = gcnew FileStream(fn, FileMode::Create);
		fs->Lock(0,fs->Length);
		BinaryWriter^ wb = gcnew BinaryWriter(fs);

		for (int i = 0; i<3; i++)
		{
			wb->Write(Header[i]);
		}
		for (int i = 0; i<T; i++)
		{
			wb->Write(I(i));
		}
		fs->Unlock(0,fs->Length);
		fs->Close();
		return true;
	}
	catch (System::Exception ^ex) { if (fs) fs->Close();  return false; }
	return false;
}

bool SaveRGBF2UC(F2UC &IR, F2UC &IG, F2UC &IB, String ^fn)
{
	FileStream^ fs;
	try
	{
		int W = IR.W;
		int H = IR.H;
		int Header[3] = { CODEREADSAVEF2UCRGB, W, H };
		int T = W*H;
		
		if (IG.W != W || IG.H != H) return false;
		if (IB.W != W || IB.H != H) return false;

		fs = gcnew FileStream(fn, FileMode::Create);
		fs->Lock(0,fs->Length);
		BinaryWriter^ wb = gcnew BinaryWriter(fs);

		for (int i = 0; i<3; i++)
		{
			wb->Write(Header[i]);
		}
		for (int i = 0; i<T; i++)
		{
			wb->Write(IR(i));
		}
		for (int i = 0; i<T; i++)
		{
			wb->Write(IG(i));
		}
		for (int i = 0; i<T; i++)
		{
			wb->Write(IB(i));
		}
		fs->Unlock(0,fs->Length);
		fs->Close();
		return true;
	}
	catch (System::Exception ^ex) { if (fs) fs->Close();  return false; }
	return false;
}


XmlNodeList ^ GetTag(XmlDocument ^doc,String ^TagName)
{
  XmlNodeList ^tag;
  try
  {
    tag=doc->GetElementsByTagName(TagName);
  }
  catch (System::Xml::XmlException ^){}
  return tag;
}

String^ GetAttribute(XmlNode ^elem,String ^AttributeName)
{
  String ^Value="";
  try
  {
		if (elem->Attributes[AttributeName])
      Value=elem->Attributes[AttributeName]->Value->ToString();
  }
  catch (System::NullReferenceException ^){}
  return Value;
}

String ^ GetStrFromTagAttribute(XmlDocument ^doc,String ^TagName, String ^AttributeName="Value")
{
  return GetAttribute(GetTag(doc,TagName)[0],AttributeName);
}

int GetIntFromStr(String ^str)
{
  int ibuf=0;
  bool ok=System::Int32::TryParse(str,ibuf);
	if (!ok) ok=System::Int32::TryParse(str,System::Globalization::NumberStyles::Any,System::Globalization::CultureInfo::CreateSpecificCulture("fr-FR"),ibuf);
	if (!ok) ok=System::Int32::TryParse(str,System::Globalization::NumberStyles::Any,System::Globalization::CultureInfo::CreateSpecificCulture("en-EN"),ibuf);
  return ibuf;
}

double GetDoubleFromStr(String ^str)
{
  double dbuf=0;
  bool ok=System::Double::TryParse(str,dbuf);
	if (!ok) ok=System::Double::TryParse(str,System::Globalization::NumberStyles::Any,System::Globalization::CultureInfo::CreateSpecificCulture("fr-FR"),dbuf);
	if (!ok) ok=System::Double::TryParse(str,System::Globalization::NumberStyles::Any,System::Globalization::CultureInfo::CreateSpecificCulture("en-EN"),dbuf);
  return dbuf;
}

int GetIntFromTagAttribute(XmlDocument ^doc, String ^TagName, String ^AttributeName = "Value")
{
	return GetIntFromStr(GetStrFromTagAttribute(doc, TagName, AttributeName));
}

double GetDoubleFromTagAttribute(XmlDocument ^doc, String ^TagName, String ^AttributeName = "Value")
{
	return GetDoubleFromStr(GetStrFromTagAttribute(doc, TagName, AttributeName));
}



int main(array<System::String ^> ^args)
{
  //Function name
  String ^FonctionName=Path::GetFileNameWithoutExtension(Environment::GetCommandLineArgs()[0]);
  ///////////////////////////////////////////////////////////////////////////////////////////////////////

  //Load xml PAR file
  String ^ fn = Environment::GetCommandLineArgs()[0]; 
  fn = fn->Substring(0, fn->LastIndexOf('\\'));
  fn=fn+"\\"+FonctionName+"PAR.xml";
  String ^line;
  StreamReader ^sr;   
  if (!File::Exists(fn))
	return -1;
  XmlDocument ^doc = gcnew XmlDocument();
  try
  {
    doc->Load(fn);
  }
  catch (System::Xml::XmlException ^e) { return -2;}

  //Data recovery from xml
  
  //input filename
  String ^fnIN=GetStrFromTagAttribute(doc,"IN");

  //output filename
  String ^fnOUT=GetStrFromTagAttribute(doc,"OUT");

  //preview filename
  String ^fnOUTPREV=GetStrFromTagAttribute(doc,"OUTPREV");

	//example of String parameter (LStr or Bool GUI type)
  String ^STRING=GetStrFromTagAttribute(doc,"strid");
  String ^BOOLEAN=GetStrFromTagAttribute(doc,"boolid");

	//example of double parameter (NUD GUI type)
  double DOUBLE=GetDoubleFromTagAttribute(doc,"doubleid");

	//example of integer parameter (NUD GUI type)
  int INTEGER=GetIntFromTagAttribute(doc,"intid");

  ///////////////////////////////////////////////////////////////////////////////////////////////////////

  //Loading image, use fnIN
	//8 bits format or RGB 8 bits format
	F2UC IM;
  F2UC IMR,IMB,IMG; //RGB
	bool loadok=true;
  bool isrgb=LoadRGBF2UC(fnIN,IMR,IMG,IMB,IM);
  if (!isrgb)
		loadok=LoadF2UC(fnIN,IM);

	//Send a message to show in plug im!, use console message
	if (loadok)
	  Console::WriteLine("Loading OK!");
	else
	{
    Console::WriteLine("Loading error...");
		return 0;
	}

	//Do something...

	//Get width and height 
  int W=GetW(IM);
	int H=GetH(IM);
  int T=W*H;


	F2UC R(W,H);
	F2UC RR(W,H);
	F2UC RG(W,H);
	F2UC RB(W,H);
	int u;

  if (BOOLEAN=="true")
	{

		//Browse all the pixels and calculate the negative image
		if (!isrgb)
		{
			for (u=0;u<T;u++)
				R(u)=255-IM(u);
		}
		else
		{
			//for RGB image, process all chanels separately
			for (u=0;u<T;u++)
			{
				RR(u)=255-IMR(u);
				RG(u)=255-IMG(u);
				RB(u)=255-IMB(u);
			}
		}

		Console::WriteLine("Negative image processed.");

		//Save a preview image to display in plug im!, use fnOUTPREV
		if (!isrgb)
			SaveF2UC(R,fnOUTPREV);
		else
			SaveRGBF2UC(RR,RG,RB,fnOUTPREV);
	}
	else
	{  
		//Another way to browse all the pixels, and now calculate horizontal gradient image
		int i,j;
		int val;
		if (!isrgb)
		{
			for (j=0;j<H;j++)
				for (i=1;i<W-1;i++)
				{
					val=IM(j,i-1)-IM(j,i+1); if (val<0) val=-val;
					R(j,i)=val;
				}
		}
		else
		{
			for (j=0;j<H;j++)
				for (i=1;i<W-1;i++)
				{
					val=IMR(j,i-1)-IMR(j,i+1); if (val<0) val=-val;
					RR(j,i)=val;
					val=IMG(j,i-1)-IMG(j,i+1); if (val<0) val=-val;
					RG(j,i)=val;
					val=IMB(j,i-1)-IMB(j,i+1); if (val<0) val=-val;
					RB(j,i)=val;
				}
		}

		Console::WriteLine("Gradient image processed.");

		//Save a new preview image to display in plug im!, use fnOUTPREV
		if (!isrgb)
			SaveF2UC(R,fnOUTPREV);
		else
			SaveRGBF2UC(RR,RG,RB,fnOUTPREV);
	}

  //Save final result, use fnOUT
	if (!isrgb)
	  SaveF2UC(R,fnOUT);
	else
    SaveRGBF2UC(RR,RG,RB,fnOUT);

  Console::WriteLine("Process completed.");

  return 0;
}
